'''
FilePath: /base-dodrio/dodrio/core/utils.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 11:51:05
LastEditors: Yixiang Chen
LastEditTime: 2025-03-24 16:56:54
'''

import os
from tqdm import tqdm

def get_file_list(inp_dir, suffix='.wav'):
    itm = []
    for home, dirs, files in os.walk(inp_dir):
        pppp = list( map(lambda fname: home + '/' + fname,
            list( filter( lambda filename: os.path.splitext(filename)[1] == suffix,
            files) ) ) )
        itm.extend(pppp)
    file_list = itm
    return file_list

def get_file_list_tail(inp_dir, suffix='_text.txt'):
    itm = []
    for home, dirs, files in os.walk(inp_dir):
        pppp = list( map(lambda fname: home + '/' + fname,
            list( filter( lambda filename: filename[-len(suffix):] == suffix,
            files) ) ) )
        itm.extend(pppp)
    file_list = itm
    return file_list


def utt_name_tran(basename, rm_prefix=False):
    #return basename
    if rm_prefix: 
        return '_'.join(basename.split('_')[1:])
    else:
        return basename

def set_wavlist(wav_dir, file_type, rm_prefix=False):
    suffix = '.'+file_type
    wavlist = get_file_list(wav_dir, suffix)
    wavdict = {}
    uttlist = []
    for wavpath in tqdm(wavlist, desc='SetList'):
        (path, filename) = os.path.split(wavpath)
        basename = filename.split(suffix)[0]
        uttname = utt_name_tran(basename, rm_prefix)
        wavdict[uttname] = wavpath
        uttlist.append(uttname)
    return wavdict, uttlist




'''
FilePath: /python-Dodrio/dodrio/core/load_info.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 15:28:08
LastEditors: Yixiang Chen
LastEditTime: 2025-03-28 11:41:23
'''

import os
import json
from tqdm import tqdm

from dodrio.core.utils import get_file_list, get_file_list_tail

################ Load info from Multi-type ######################

def load_info_dict_emilia(info_dir, kl):
    json_list = get_file_list(info_dir, '.json')
    keys_list = ["id", "wav", "text", "duration", "speaker", "language", "dnsmos"]
    info_dict = {}
    for jsonf in tqdm(json_list, desc='LoadJson'):
        with open(jsonf, 'r') as ojf:
            data = json.load(ojf)
        info_dict[data["id"]] = data
    return info_dict, keys_list

def load_info_dict_tfile_supdir(info_dir, kl):
    sufdict = {'text': '_text.txt', 'punc_text': '_text_punc.txt'}
    titleid_list = os.listdir(info_dir)
    if 'data.lst' in titleid_list: 
        titleid_list.remove('data.lst')
    info_dict = {}
    punc_flag = ('punc_text' in kl)
    for title in tqdm(titleid_list, desc='LoadInfo4Supdir'):
        title_dir = os.path.join(info_dir, title)
        spk_list = os.listdir(title_dir)
        for spk in spk_list:
            spk_dir = os.path.join(title_dir, spk)
            tf_list = get_file_list_tail(spk_dir, sufdict['text'])
            for tff in tf_list:
                #basename = os.path.split(tff)[-1].split('.')[0]
                uttid = os.path.split(tff)[-1].split(sufdict['text'])[0]
                with open(tff, 'r') as otff:
                    text = otff.read().strip()
                if not punc_flag:
                    useinfo = {'id':uttid, 'titleid':title, 'speaker':spk, 'text':text}
                else:
                    ptf = os.path.join(spk_dir, uttid + sufdict['punc_text'])
                    if os.path.isfile(ptf):
                        with open(ptf, 'r') as optf: 
                            punc_text = optf.read().strip()
                    else:
                        punc_text = None
                    useinfo = {'id':uttid, 'titleid':title, 'speaker':spk, 'text':text, 'punc_text':punc_text}
                info_dict[uttid] = useinfo
    if punc_flag:
        keys_list = ['id', 'titleid', 'speaker', 'text', 'punc_text']
    else:
        keys_list = ['id', 'titleid', 'speaker', 'text']
    return info_dict, keys_list
            
def load_info_dict_genshin(info_dir, kl=['text']):
    sufdict = {'text': '.lab'}
    info_dict = {}
    spk_list = os.listdir(info_dir)
    for spk in spk_list:
        spk_dir = os.path.join(info_dir, spk)
        tf_list = get_file_list(spk_dir, sufdict['text'])
        for tff in tf_list:
            #basename = os.path.split(tff)[-1].split('.')[0]
            uttid = os.path.split(tff)[-1].split(sufdict['text'])[0]
            with open(tff, 'r') as otff:
                text = otff.read().strip()
            useinfo = {'id':uttid, 'speaker':spk, 'text':text, 'language':'ZH'}
            info_dict[uttid] = useinfo
    keys_list = ['id', 'speaker', 'text', 'language']
    return info_dict, keys_list

def load_info_dict_single_spk(info_dir, kl=['text']):
    sufdict = {'text': '_textpuncnew.txt'}
    info_dict = {}
    spk = os.path.split(info_dir)[-1] 
    spk_dir = info_dir
    tf_list = get_file_list_tail(spk_dir, sufdict['text'])
    for tff in tf_list:
        #basename = os.path.split(tff)[-1].split('.')[0]
        uttid = os.path.split(tff)[-1].split(sufdict['text'])[0]
        with open(tff, 'r') as otff:
            text = otff.read().strip()
        useinfo = {'id':uttid, 'speaker':spk, 'text':text, 'language':'ZH'}
        info_dict[uttid] = useinfo
    keys_list = ['id', 'speaker', 'text', 'language']
    return info_dict, keys_list

def load_info_dict_table(info_dir, kl):
    use_file = 'info_align.list'
    titleid_list = os.listdir(info_dir)
    if 'data.lst' in titleid_list: 
        titleid_list.remove('data.lst')
    info_dict = {}
    for title in tqdm(titleid_list, desc='LoadInfo4Supdir'):
        titleid = title.split('_')[0]
        title_dir = os.path.join(info_dir, title)
        usefile_path = os.path.join(title_dir, use_file)
        with open(usefile_path, 'r') as ouf:
            lines = ouf.readlines()
        for line in lines:
            spl = line.strip().split('|')
            wavp, spk, lang, text, phoneme, dur = spl[0], spl[1], spl[2], spl[3], spl[4], spl[5]
            uttid = os.path.split(wavp)[-1].split('.')[0]
            useinfo = {'id':uttid, 'titleid':titleid, 'speaker':spk, 'text':text, 'language':lang}
            info_dict[uttid] = useinfo
    keys_list = ['id', 'titleid', 'speaker', 'text', 'language']
    return info_dict, keys_list

def load_info_dict_libritts(info_dir, kl):
    sufdict = {'text': '.normalized.txt', 'unnorm_text': '.original.txt'}
    spkid_list = os.listdir(info_dir)
    info_dict = {}
    unnorm_flag = ('unnorm_text' in kl)
    for spk in tqdm(spkid_list, desc='LoadInfo4Supdir'):
        spk_dir = os.path.join(info_dir, spk)
        chapter_list = os.listdir(spk_dir)
        for chapter in chapter_list:
            chapter_dir = os.path.join(spk_dir, chapter)
            tf_list = get_file_list_tail(chapter_dir, sufdict['text'])
            for tff in tf_list:
                #basename = os.path.split(tff)[-1].split('.')[0]
                uttid = os.path.split(tff)[-1].split(sufdict['text'])[0]
                with open(tff, 'r') as otff:
                    text = otff.read().strip()
                if not unnorm_flag:
                    useinfo = {'id':uttid, 'chapterid':chapter, 'speaker':spk, 'text':text}
                else:
                    ptf = os.path.join(chapter_dir, uttid + sufdict['unnorm_text'])
                    if os.path.isfile(ptf):
                        with open(ptf, 'r') as optf: 
                            unnorm_text = optf.read().strip()
                    else:
                        unnorm_text = None
                    useinfo = {'id':uttid, 'chapterid':chapter, 'speaker':spk, 'text':text, 'unnorm_text':unnorm_text}
                info_dict[uttid] = useinfo
    if unnorm_flag:
        keys_list = ['id', 'chapterid', 'speaker', 'text', 'unnorm_text']
    else:
        keys_list = ['id', 'chapterid', 'speaker', 'text']
    return info_dict, keys_list





'''
FilePath: /python-Dodrio/dodrio/afeat/exp_fun.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-27 09:59:30
LastEditors: Yixiang Chen
LastEditTime: 2025-04-29 17:33:45
'''

import numpy as np
import torch
import torchaudio
import torchaudio.compliance.kaldi as kaldi
import onnxruntime

import logging

import whisper

class extractor_embedding:
    def __init__(self, onnx_path) -> None:
        option = onnxruntime.SessionOptions()
        option.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        option.intra_op_num_threads = 1
        providers = ["CPUExecutionProvider"]
        self.ort_session = onnxruntime.InferenceSession(onnx_path, sess_options=option, providers=providers)
        
        self.spk2embedding = {}
    
    def extractor(self, floatnpwav, utt, utt2spk, sample_rate=48000):
        audio = torch.from_numpy(floatnpwav)
        audio = audio.float().unsqueeze(0)
        if sample_rate != 16000:
            audio = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)(audio)
        feat = kaldi.fbank(audio,
                           num_mel_bins=80,
                           dither=0,
                           sample_frequency=16000)
        feat = feat - feat.mean(dim=0, keepdim=True)
        embedding = self.ort_session.run(None, {self.ort_session.get_inputs()[0].name: feat.unsqueeze(dim=0).cpu().numpy()})[0].flatten().tolist()
        if utt in utt2spk.keys():
            spk = utt2spk[utt]
        else:
            spk = 'SpkNone'
        if spk not in self.spk2embedding:
            self.spk2embedding[spk] = []
        self.spk2embedding[spk].append(embedding)
        embedding = np.array(embedding)
        return embedding
    
    def mean_spk_embedding(self):
        for k, v in self.spk2embedding.items():
            self.spk2embedding[k] = torch.tensor(v).mean(dim=0).tolist()
    
    def spk_embedding_save(self, floatnpwav, utt, utt2spk):
        if utt in utt2spk.keys():
            spk = utt2spk[utt]
        else:
            spk = 'SpkNone'
        spkembedding = self.spk2embedding[spk]
        spkembedding = np.array(spkembedding)
        return spkembedding

class speech_token_extractor:
    def __init__(self, onnx_path):
        option = onnxruntime.SessionOptions()
        option.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        option.intra_op_num_threads = 1
        providers = ["CUDAExecutionProvider"]
        self.ort_session = onnxruntime.InferenceSession(onnx_path, sess_options=option, providers=providers)
    
    def extractor(self, floatnpwav, utt, sample_rate=48000):
        audio = torch.from_numpy(floatnpwav)
        audio = audio.float().unsqueeze(0)
        if sample_rate != 16000:
            audio = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=16000)(audio)
        if audio.shape[1] / 16000 > 30:
            logging.warning('do not support extract speech token for audio longer than 30s')
            speech_token = None
        else:
            feat = whisper.log_mel_spectrogram(audio, n_mels=128)
            speech_token = self.ort_session.run(None, {self.ort_session.get_inputs()[0].name: feat.detach().cpu().numpy(),
                                                  self.ort_session.get_inputs()[1].name: np.array([feat.shape[2]], dtype=np.int32)})[0].flatten().tolist()
            speech_token = np.array(speech_token)
        return speech_token


from modelscope.pipelines import pipeline
from modelscope.utils.constant import Tasks
import librosa

import os
import sys
class suppress_stdout_stderr(object):
    def __enter__(self):
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = open(os.devnull, 'w')
    def __exit__(self, *_):
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr

class emotion2vec_extractor:
    def __init__(self):
        self.inference_pipeline = pipeline(
            task=Tasks.emotion_recognition,
            model="iic/emotion2vec_base_finetuned",
            disable_update=True)
    
    def extractor(self, floatnpwav, utt, sample_rate=48000):
        floatnpwav = librosa.resample(floatnpwav, orig_sr=sample_rate, target_sr=16000)
        #with warnings.catch_warnings():
        #    warnings.simplefilter('ignore')
        with suppress_stdout_stderr():
            rec_result = self.inference_pipeline(input=floatnpwav.astype(np.float32), granularity="utterance", extract_embedding=True)
        scores = rec_result[0]['scores']
        embedding = rec_result[0]['feats']
        str_scores = [str(xx) for xx in scores]
        str_scores = ','.join(str_scores)
        return embedding, [str_scores]




'''
FilePath: /base-dodrio/dodrio/core/parquet_base.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 11:26:49
LastEditors: Yixiang Chen
LastEditTime: 2025-03-24 16:57:57
'''

import os
import pandas as pd
from tqdm import tqdm
import math
import multiprocessing
import glob
from scipy.io import wavfile
import pyarrow.parquet as pq

from tinytag import TinyTag # for read mp3 metainfo

from dodrio.core.utils import set_wavlist

############## Wav to Parquet ###############

def save_parquet(wavinfo_dict, savelist, parquet_fn):
    sr_list = [wavinfo_dict[x][0] for x in savelist]
    dtype_list = [wavinfo_dict[x][1] for x in savelist]
    audio_list = [wavinfo_dict[x][2] for x in savelist]
    df = pd.DataFrame()
    df['utt'] = savelist
    df['sample_rate'] = sr_list
    df['dtype'] = dtype_list
    df['audio_data'] = audio_list
    df.to_parquet(parquet_fn)
    print(f"{parquet_fn} had be saved")

def get_mp3_metainfo(mp3file):
    tag = TinyTag.get(mp3file)
    return tag.samplerate

def gen_parquet(wav_dir, parquet_dir, mid_name='', file_type='wav', num_utts_per_parquet=2000, num_processes=5, process_max_num=10000):
    os.makedirs(parquet_dir, exist_ok=True)
    wavdict, uttlist = set_wavlist(wav_dir, file_type)

    parquet2utt = {}
    turn_num = math.ceil(len(uttlist) / process_max_num)
    for tid in range(turn_num):
        wavinfo_dict ={}
        for utt in tqdm(uttlist[process_max_num * tid : process_max_num * (tid+1)], desc=f'Turn{str(tid)}LoadAudio'):
            wavpath = wavdict[utt]
            if file_type == 'wav':
                sr, wav = wavfile.read(wavpath)
                if len(wav)<1:
                    uttlist.remove(utt)
                    print(f"{utt} wavfile is None")
                    continue
                if len(wav.shape) > 1:
                    print(f"{utt} {str(len(wav.shape))} channel is non-mono channel, just first channel will be save")
                    wav = wav[:,0]
                dtype = str(wav.dtype)
                wavinfo_dict[utt] = [sr, dtype, wav]
            elif file_type == 'mp3':
                byte_mp3_data = open(wavpath, 'rb').read()
                sr = get_mp3_metainfo(wavpath)
                dtype = 'mp3'
                wavinfo_dict[utt] = [sr, dtype, byte_mp3_data]
            else:
                print("Now just accept mp3 and wav format")
                return

        # Using process pool to speedup
        prefix = file_type
        pool = multiprocessing.Pool(processes=num_processes)
        parquet_list = []
        for i, j in enumerate(range(process_max_num * tid, min(len(uttlist), process_max_num * (tid+1)), num_utts_per_parquet)):
            idx = (process_max_num * tid) // num_utts_per_parquet + i
            pfile = prefix + '_' + mid_name + '_{:05d}.parquet'.format(idx)
            parquet_file = os.path.join(parquet_dir, pfile)
            parquet_list.append(parquet_file)
            parquet2utt[pfile] = uttlist[j: j + num_utts_per_parquet] 
            pool.apply_async(save_parquet, (wavinfo_dict, uttlist[j: j + num_utts_per_parquet], parquet_file))
        pool.close()
        pool.join()

    utt2parquet_file = os.path.join(parquet_dir, 'utt2parquet.list') 
    u2pf = open(utt2parquet_file, 'w')
    for pak in parquet2utt.keys():
        for utt in parquet2utt[pak]:
            outline = utt+'|' + pak + '\n'
            u2pf.write(outline)
    u2pf.close()


############## Parquet to Wav ###############

def parquet2wav_single(parquet_file, wav_dir): 
    ftype = os.path.split(parquet_file)[-1].split('_')[0]
    df = pq.read_table(parquet_file).to_pandas()
    basename = os.path.split(parquet_file)[-1].split('.parquet')[0]
    for idx in tqdm(range(len(df)), desc=f'{basename} Processing'):
        utt = df.iloc[idx]['utt']
        sr = df.iloc[idx]['sample_rate']
        dtype = df.iloc[idx]['dtype']
        audio = df.iloc[idx]['audio_data']
        if ftype == 'wav':
            wavpath = os.path.join(wav_dir, utt+'.wav')
            wavfile.write(wavpath, sr, audio)
        else:
            wavpath = os.path.join(wav_dir, utt+'.'+ftype)
            with open(wavpath, 'wb') as ww:
                ww.write(audio)

def parquet2wav(parquet_dir, wav_dir):
    os.makedirs(wav_dir, exist_ok=True)
    pq_file_list = glob.glob(parquet_dir+'/*.parquet')
    for parquet_file in pq_file_list:
        parquet2wav_single(parquet_file, wav_dir)


'''
FilePath: /python-Dodrio/dodrio/tools/load_data.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-26 19:23:27
LastEditors: Yixiang Chen
LastEditTime: 2025-04-21 16:09:43
'''


import os
from io import BytesIO
import librosa
from tqdm import tqdm
import numpy as np
import pyarrow.parquet as pq


####################### Load data block ###################### 

def get_file_list(inp_dir, suffix='.wav'):
    itm = []
    for home, dirs, files in os.walk(inp_dir):
        pppp = list( map(lambda fname: home + '/' + fname,
            list( filter( lambda filename: os.path.splitext(filename)[1] == suffix,
            files) ) ) )
        itm.extend(pppp)
    file_list = itm
    return file_list

def load_datalist_dict(input_dir, from_type='parquet'):
    if from_type == 'parquet':
        load_file = os.path.join(input_dir, 'utt2parquet.list')
    else:
        load_file = os.path.join(input_dir, 'uttinfo.list')
    data_dict = {}
    with open(load_file, 'r') as lf:
        lines = lf.readlines()
    for line in lines:
        spl = line.strip().split('|')
        name= spl[0]
        packname = spl[1].split('.')[0]
        data_dict.setdefault(packname, []).append(name)
    return data_dict

def load_data_dict(input_dir, from_type='parquet'):
    if from_type=='parquet':
        suffix = '.parquet'
        infolistf = ''
    else:
        suffix = '.pack'
        infolistf = os.path.join(input_dir, 'uttinfo.list')
    packlist = get_file_list(input_dir, suffix)
    packlist.sort()
    data_dict = {}
    for idx in range(len(packlist)):
        packfile = packlist[idx]
        basename = os.path.split(packfile)[-1].split(suffix)[0]
        uttdict = load_pack_audio_data(packfile, infolistf, return_sr = False)
        data_dict[basename] = uttdict
    return data_dict


####################### Load audio ######################  

def audiotran(data, atype, dtype):
    if atype == 'wav':
        if dtype == 'int16':
            dnum = 32768
        elif dtype == 'int32':
            dnum = 2147483648
        else:
            print(f"{dtype} is not normal data type")
            return 
        return data / dnum
    elif atype == 'mp3':
        wav, sr = librosa.load(BytesIO(data))
        return wav

def load_pack_audio_data(packp, infolistf='', return_sr = False):
    ftype = os.path.split(packp)[-1].split('.')[-1]
    assert ftype in ['parquet', 'pack']
    audio_type = os.path.split(packp)[-1].split('_')[0]
    assert audio_type in ['wav', 'mp3']
    outdict = {}
    if ftype == 'parquet':
        df = pq.read_table(packp).to_pandas()
        for idx in tqdm(range(len(df)), desc=f'Loading'):
            utt = df.iloc[idx]['utt']
            sr = df.iloc[idx]['sample_rate']
            dtype = df.iloc[idx]['dtype']
            audio = df.iloc[idx]['audio_data']
            wav = audiotran(audio, audio_type, dtype)
            if return_sr:
                outdict[utt] = [wav,sr]
            else:
                outdict[utt] = wav
    else:
        info_file = os.path.join(infolistf)
        with open(info_file, 'r') as iff:
            info_list = iff.readlines()
        packname = os.path.split(packp)[-1] 
        #for info in tqdm(info_list):
        for info in info_list:
            utt, pf, start, end = info.split('|')
            start = int(start)
            end = int(end)
            if pf != packname:
                continue
            with open(packp, 'rb') as opf :
                opf.seek(start)
                data = opf.read(end-start)
            wav = np.frombuffer(data, dtype=np.int16)
            wav = wav / 32768
            if return_sr:
                outdict[utt] = [wav,48000]
            else:
                outdict[utt] = wav
    return outdict


####################### Load single ######################

def load_audio_single(packp, start, end):
    with open(packp, 'rb') as opf :
        opf.seek(start)
        data = opf.read(end-start)
    wav = np.frombuffer(data, dtype=np.int16)
    wav = wav / 32768
    return wav

def load_feat_single(featpack, start, end, shape):
    with open(featpack, 'rb') as opf :
        opf.seek(start)
        data = opf.read(end-start)
    feat = np.frombuffer(data, dtype=np.float32)
    feat = np.reshape(feat, shape)
    return feat

def load_data_from_line(infoline):
    spl = infoline.strip().split('|')
    data_dict = {}
    uttid = spl[0]
    wavp, wstart, wend = spl[1], int(spl[2]), int(spl[3])
    audio = load_audio_single(wavp, wstart, wend) 
    spkid, text, language = spl[4], spl[5], spl[6]

    data_dict['uttid'] = uttid
    data_dict['audio'] = audio
    data_dict['spkid'] = spkid
    data_dict['text'] = text
    data_dict['language'] = language

    feat_type_num = (len(spl)- 7) // 5
    for idx in range(feat_type_num):
        beginid = 7 + idx*5 
        featname = spl[beginid]
        featpack = spl[beginid+1]
        fstart, fend, fshape = int(spl[beginid+2]), int(spl[beginid+3]), spl[beginid+4]
        fshape = [int(xx) for xx in fshape.split(',')]
        feat_data = load_feat_single(featpack, fstart, fend, fshape)
        data_dict[featname] = feat_data

    return data_dict 
        
def load_data_from_line_align(infoline):
    spl = infoline.strip().split('|')
    data_dict = {}
    uttid = spl[0]
    wavp, wstart, wend = spl[1], int(spl[2]), int(spl[3])
    audio = load_audio_single(wavp, wstart, wend) 
    spkid, text, language = spl[4], spl[5], spl[6]

    phones, duration = spl[7], spl[8]
    phones = phones.split()
    duration = [int(xx) for xx in duration.split()]

    data_dict['uttid'] = uttid
    data_dict['audio'] = audio
    data_dict['spkid'] = spkid
    data_dict['text'] = text
    data_dict['language'] = language
    data_dict['phones'] = phones
    data_dict['duration'] = duration

    feat_type_num = (len(spl)- 9) // 5
    for idx in range(feat_type_num):
        beginid = 9 + idx*5 
        featname = spl[beginid]
        featpack = spl[beginid+1]
        fstart, fend, fshape = int(spl[beginid+2]), int(spl[beginid+3]), spl[beginid+4]
        fshape = [int(xx) for xx in fshape.split(',')]
        feat_data = load_feat_single(featpack, fstart, fend, fshape)
        data_dict[featname] = feat_data

    return data_dict 





'''
FilePath: /python-Dodrio/dodrio/afeat/exp_audiof.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-04-09 14:56:10
LastEditors: Yixiang Chen
LastEditTime: 2025-04-24 15:47:34
'''

from librosa.filters import mel as librosa_mel_fn
import torch 
import torch.nn.functional as F
import numpy as np
from functools import partial

# for pitch
import pyworld as pw

mel_basis = {}
hann_window = {}

def dynamic_range_compression_torch(x, C=1, clip_val=1e-5):
    return torch.log(torch.clamp(x, min=clip_val) * C)

def spectral_normalize_torch(magnitudes):
    output = dynamic_range_compression_torch(magnitudes)
    return output

def mel_extractor_torch(wav, n_fft, num_mels, sampling_rate, hop_size, win_size, fmin, fmax, center=False):
    '''
        wav shape : (B, sample_len) torch.Tensor, range [-1, 1], Float 32
        spec out shape : (B, mel_dim, spec_len) torch.Tensor, Float 32
        Float 32  because mel_basis by librosa
    '''
    if torch.min(wav) < -1.0:
        print("min value is ", torch.min(wav))
    if torch.max(wav) > 1.0:
        print("max value is ", torch.max(wav))

    # Get Mel Wrap Matix
    global mel_basis, hann_window  # pylint: disable=global-statement
    if f"{str(fmax)}_{str(wav.device)}" not in mel_basis:
        mel = librosa_mel_fn(sr=sampling_rate, n_fft=n_fft, n_mels=num_mels, fmin=fmin, fmax=fmax)
        mel_basis[str(fmax) + "_" + str(wav.device)] = torch.from_numpy(mel).float().to(wav.device)
        hann_window[str(wav.device)] = torch.hann_window(win_size).to(wav.device)

    if wav.shape[-1] > int((n_fft - hop_size) / 2):
        wav_pad = torch.nn.functional.pad(
            wav.unsqueeze(1), (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)), mode="reflect"
        )
    else:
        wav_pad = torch.nn.functional.pad(
            wav.unsqueeze(1), (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)), mode='constant'
        )

    wav_pad = wav_pad.squeeze(1)

    spec = torch.view_as_real(
        torch.stft(
            wav_pad,
            n_fft,
            hop_length=hop_size,
            win_length=win_size,
            window=hann_window[str(wav_pad.device)],
            center=center,
            pad_mode="reflect",
            normalized=False,
            onesided=True,
            return_complex=True,
        )
    ) #  [B, nfft//2+1, seq_len, 2]

    #phase = torch.atan2(spec[...,1], spec[...,0])

    spec = torch.sqrt(spec.pow(2).sum(-1) + (1e-9))
    magnitudes = spec # [B, nfft//2+1, seq_len]

    spec = torch.matmul(mel_basis[str(fmax) + "_" + str(wav_pad.device)], spec)
    # default 1e-5 clip norm
    spec = spectral_normalize_torch(spec)

    return spec, magnitudes


def pitch_extract_pyworld(wav, mel_len, sample_rate, hop_size, etype='harvest'):
    '''
        wav shape : [1, sample_len] torch.Tensor, range [-1, 1]
        out f0 shape : [seq_len] torch.Tensor 
    '''
    frame_period = hop_size * 1000 / sample_rate
    if etype == 'harvest':
        _f0, t = pw.harvest(wav.squeeze(dim=0).numpy().astype('double'), sample_rate, frame_period=frame_period)
    else:
        _f0, t = pw.dio(wav.squeeze(dim=0).numpy().astype('double'), sample_rate, frame_period=frame_period)
    if sum(_f0 != 0) < 5: # this happens when the algorithm fails
        _f0, t = pw.dio(wav.squeeze(dim=0).numpy().astype('double'), sample_rate, frame_period=frame_period) # if harvest fails, try dio
    f0 = pw.stonemask(wav.squeeze(dim=0).numpy().astype('double'), _f0, t, sample_rate)
    if mel_len > 1:
        f0 = F.interpolate(torch.from_numpy(f0).view(1, 1, -1), size=mel_len, mode='linear').view(-1)
    return f0

def energy_extract_time(wav, hop_size, win_size):
    # wav shape : [1, sample_len] torch.Tensor, range [-1, 1]
    # out shape : [seq_len] torch.Tensor 
    feat_len = wav.shape[-1] // hop_size
    energy_list = []
    for idx in range(feat_len):
        piece = wav[:,idx*hop_size:(idx+1)*hop_size]
        energy = torch.sum(piece**2)
        energy_list.append(energy)
    return torch.tensor(energy_list)

def energy_extract_spec(magnitudes):
    # magnitudes [B, nfft//2+1, seq_len]
    # magnitudes 2-norm on seq_len dim 
    energy = torch.norm(magnitudes, dim=1)
    return energy
        


pitch_fun_dict = {
    'pyworld': pitch_extract_pyworld,
}
mel_fun_dict = {
    'torch': mel_extractor_torch,
}

def extract_mfe(floatnpwav, utt, meltype='torch', f0type='pyworld', energytype='spec',
                sampling_rate=48000, hop_size=512, 
                n_fft=2048, num_mels=80, win_size=2048, fmin=0, fmax=8000, mel_center=False,
                f0etype='harvest',
                ):
    '''
        wav shape : (B, sample_len) torch.Tensor, range [-1, 1], Float 32
        f0 shape : [seq_len] torch.Tensor 
        energy shape : [seq_len] torch.Tensor 
        mel_spec out shape : (B, mel_dim, spec_len) torch.Tensor, Float 32
    '''

    wav = torch.from_numpy(floatnpwav)
    wav = wav.float().unsqueeze(0)
    if wav.shape[-1] < n_fft:
        wav = torch.nn.functional.pad(
            wav, (n_fft//2 +1, n_fft//2+1), mode='constant'
        )

    pitch_extractor = partial(pitch_fun_dict[f0type], 
                                sample_rate=sampling_rate, hop_size=hop_size,
                                etype=f0etype)
    mel_extractor = partial(mel_fun_dict[meltype],
                                n_fft=n_fft, num_mels=num_mels, 
                                sampling_rate=sampling_rate, hop_size=hop_size, win_size=win_size, 
                                fmin=fmin, fmax=fmax, center=mel_center) 
    
    mel_spec, magnitudes = mel_extractor(wav) 
    pitch = pitch_extractor(wav, mel_spec.shape[-1]) 

    if energytype == 'time':
        energy_extractor = partial(energy_extract_time, 
                                    hop_size=hop_size, win_size=win_size)
        energy = energy_extractor(wav)
        if mel_spec.shape[-1]-energy.shape[-1] > 0:
            energy = F.pad(energy, (0, mel_spec.shape[-1]-energy.shape[-1]))
        energy = energy[...,:mel_spec.shape[-1]]
    else:
        energy_extractor = energy_extract_spec
        energy = energy_extractor(magnitudes)
    
    pitch = pitch.squeeze(0).cpu().numpy().astype(np.float32)
    energy = energy.squeeze(0).cpu().numpy().astype(np.float32)
    mel_spec = mel_spec.squeeze(0).cpu().numpy().astype(np.float32)

    return pitch, energy, mel_spec

################# Compute mean std ###############

from dodrio.tools.load_data import load_feat_single
from sklearn.preprocessing import StandardScaler
import json
from tqdm import tqdm

def load_feat_from_line(line, featname):
    spl = line.strip().split('|')
    uidx = -1
    for idx in range(len(spl)):
        if spl[idx] == featname:
            uidx = idx
            break
    if uidx == -1:
        print("There is no ", featname)
        return 0
    fshape = [int(xx) for xx in spl[uidx+4].split(',')]
    feat = load_feat_single(spl[uidx+1], int(spl[uidx+2]), int(spl[uidx+3]), fshape)
    return feat

def getStMag(inpfile, featname):
    feat_scaler = StandardScaler()
    with open(inpfile, 'r') as oif:
        lines = oif.readlines()
    for line in tqdm(lines):
        feat = load_feat_from_line(line, featname)
        if len(feat) > 0:
            feat_scaler.partial_fit(feat.reshape((-1, 1)))
    try:
        feat_mean = feat_scaler.mean_[0]
        feat_std = feat_scaler.scale_[0]
    except:
        feat_mean = 0
        feat_std = 1
    return feat_mean, feat_std

def saveStMag(inpfile, outfile, featname):
    feat_mean, feat_std = getStMag(inpfile, featname)
    outdict = {
        "mean": feat_mean,
        "std": feat_std 
    }
    with open(outfile, "w", encoding="utf-8") as f:
        json.dump(outdict, f, indent=2, ensure_ascii=False)







'''
FilePath: /python-Dodrio/dodrio/genlist/gendatalist.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-02-08 17:25:21
LastEditors: Yixiang Chen
LastEditTime: 2025-04-15 16:58:48
'''

import os
import json

LANGUAGE_NONE = 'LangNone'

def checkutt(utt, pack_id, start, end, spk, text, language, ptype='pack'):
    threshold_audio = 10 # 48000 sample rate 
    if ptype == 'pack':
        if (int(end) - int(start)) > threshold_audio * 2:
            return True
    else:
        return False

def SetProcess_pack(uttinfo_text_file, package_info_file, package_dir):
    with open(uttinfo_text_file, 'r') as otif:
        ti_lines = otif.readlines()
    info_dict = {}
    index_list = ti_lines[0].strip().split('|')
    spkid, textid, langid = -1, -1, -1 
    for idx in range(len(index_list)):
        if index_list[idx] == 'speaker':
            spkid = idx
        elif index_list[idx] == 'text':
            textid = idx
        if 'language' in index_list:
            if index_list[idx] == 'language': 
                langid = idx
    for idx in range(1, len(ti_lines)):
        spl = ti_lines[idx].strip().split('|')
        utt = spl[0]
        spk = spl[spkid]
        text = spl[textid]
        language = spl[langid] if langid != -1 else LANGUAGE_NONE
        info_dict[utt] = [spk, text, language]
    del ti_lines
    with open(package_info_file, 'r') as opif:
        p_lines = opif.readlines()
    outinfo_list = []
    for idx in range(len(p_lines)):
        spl = p_lines[idx].strip().split('|')
        utt, pack_id, start, end = spl[0], spl[1], spl[2], spl[3]
        # utt_id|save_type|pack_path|start_pointer|end_pointer|spk|language|text
        spk, text, language = info_dict[utt]
        if checkutt(utt, pack_id, start, end, spk, text, language, ptype='pack'):
            outline = [utt, 'pack', os.path.join(package_dir, pack_id), start, end, spk, language, text]
            outinfo_list.append(outline)
    return outinfo_list

def SetProcess_dir(package_dir, info_dir):
    uttinfo_text_file = os.path.join(info_dir, 'uttinfo_text.list')
    package_info_file = os.path.join(package_dir, 'uttinfo.list')
    use_info_list = SetProcess_pack(uttinfo_text_file, package_info_file, package_dir)
    return use_info_list

def genListDir(supdir_list, outdir, prefix='test', subnum=50000):

    os.makedirs(outdir, exist_ok=True)
    all_list_file = os.path.join(outdir, 'all_usage_utt.list')
    sub_table_dir = os.path.join(outdir, 'subtable')
    os.makedirs(sub_table_dir, exist_ok=True)
    spk_id_file = os.path.join(outdir, 'spk_name.dict') 
    list_list_file = os.path.join(outdir, 'list_subtable.list')

    all_info_list = []
    for supdir in supdir_list:
        package_dir = os.path.join(supdir, 'pack_dir') 
        info_dir = os.path.join(supdir, 'info_dir')
        use_info_list = SetProcess_dir(package_dir, info_dir)
        all_info_list.extend(use_info_list)
    
    spkdict = {}
    spkid = 0
    save_all = True
    olistsub = open(list_list_file, 'w')
    if save_all:
        oallt = open(all_list_file, 'w')
    for idx in range(0, len(all_info_list), subnum):
        use_id = (idx+1) // subnum 
        subtable_file = os.path.join(sub_table_dir, prefix+'_sub_'+str(use_id).rjust(6,'0')+ '.table') 
        olistsub.write(subtable_file+'\n')
        osubt = open(subtable_file, 'w')
        for ik in range(idx, min(len(all_info_list), idx+subnum)):
            info_line = all_info_list[ik]
            spk = info_line[5]
            if spk not in spkdict.keys():
                spkdict[spk] = spkid
                spkid += 1
            outline = '|'.join(info_line) + '\n'
            osubt.write(outline)
            if save_all:
                oallt.write(outline)
        osubt.close()
    if save_all:
        oallt.close()
    olistsub.close()
    with open(spk_id_file, 'w', encoding="utf-8") as opf:
        json.dump(spkdict, opf, indent=2, ensure_ascii=False)
    
########################  new version  ###################################

def pack_dict_load(package_dir):
    package_info_file = os.path.join(package_dir, 'uttinfo.list')
    with open(package_info_file, 'r') as opif:
        p_lines = opif.readlines()
    outinfo_dict = {}
    for idx in range(len(p_lines)):
        spl = p_lines[idx].strip().split('|')
        utt, pack_id, start, end = spl[0], spl[1], spl[2], spl[3]
        packpath = os.path.join(package_dir, pack_id)
        outinfo_dict[utt] = [packpath, start, end]
    keylist = ['packpath', 'start', 'end']
    return outinfo_dict, keylist 

def info_dict_load(info_dir):
    uttinfo_text_file = os.path.join(info_dir, 'uttinfo_text.list')
    with open(uttinfo_text_file, 'r') as otif:
        ti_lines = otif.readlines()
    info_dict = {}
    index_list = ti_lines[0].strip().split('|')
    spkid, textid, langid = -1, -1, -1 
    for idx in range(len(index_list)):
        if index_list[idx] == 'speaker':
            spkid = idx
        elif index_list[idx] == 'text':
            textid = idx
        if 'language' in index_list:
            if index_list[idx] == 'language': 
                langid = idx
    for idx in range(1, len(ti_lines)):
        spl = ti_lines[idx].strip().split('|')
        utt = spl[0]
        spk = spl[spkid]
        text = spl[textid]
        language = spl[langid] if langid != -1 else LANGUAGE_NONE
        info_dict[utt] = [spk, text, language]
    keylist = ['speaker', 'text', 'language']
    del ti_lines 
    return info_dict, keylist

def feat_dict_load(feat_dir, featname):
    featinfo_file = os.path.join(feat_dir, 'feat_info_'+ featname +'.list')
    with open(featinfo_file, 'r') as ofif:
        info_lines = ofif.readlines()
    info_dict = {}
    for line in info_lines:
        spl = line.strip().split('|')
        utt, packid, start, end, shape = spl[0], spl[1], spl[2], spl[3], spl[4]
        packpath = os.path.join(feat_dir, packid)
        info_dict[utt] = [packpath, start, end, shape]
    keylist = ['packpath', 'start', 'end', 'shape' ]
    return info_dict, keylist 

def align_dict_load(align_dir, featname='align'):
    featinfo_file = os.path.join(align_dir, 'feat_info_'+ featname +'.list')
    with open(featinfo_file, 'r') as ofif:
        info_lines = ofif.readlines()
    info_dict = {}
    for line in info_lines:
        spl = line.strip().split('|')
        utt, packid, phone_list, phone_dur = spl[0], spl[1], spl[2], spl[3]
        info_dict[utt] = [phone_list, phone_dur]
    keylist = ['phones', 'duration']
    return info_dict, keylist 

def datadirProcess(datadir, featlist, check_func):
    all_keylist = []
    package_dir = os.path.join(datadir, 'pack_dir')
    wav_info_dict, kl1 = pack_dict_load(package_dir)
    all_keylist.extend(kl1)
    info_dir = os.path.join(datadir, 'info_dir')
    info_dict, kl2 = info_dict_load(info_dir)
    all_keylist.extend(kl2)
    supfeat_dict = {}
    kl_dict = {}
    for featname in featlist:
        feat_dir = os.path.join(datadir, featname+'_dir')
        supfeat_dict[featname], kl_dict[featname] =  feat_dict_load(feat_dir, featname)
        all_keylist.append('featname_'+featname)
        all_keylist.extend(kl_dict[featname])
    
    out_dict = {}
    intersection = wav_info_dict.keys() & info_dict.keys() 
    for featname in featlist:
        intersection = intersection & supfeat_dict[featname].keys() 
    
    for utt in intersection:
        outinfo_list = wav_info_dict[utt]
        outinfo_list.extend(info_dict[utt])
        for featname in featlist:
            outinfo_list.append(featname)
            outinfo_list.extend(supfeat_dict[featname][utt])
        if check_func(outinfo_list):
            out_dict[utt] = outinfo_list
    return out_dict, all_keylist

def datadirProcess_align(datadir, featlist, check_func):
    all_keylist = []
    package_dir = os.path.join(datadir, 'pack_dir')
    wav_info_dict, kl1 = pack_dict_load(package_dir)
    all_keylist.extend(kl1)
   
    align_flag = False
    if 'align' in featlist:
        align_flag = True
        featlist.remove('align')
        align_dir = os.path.join(datadir, 'align_dir')
        align_dict, kl3 = align_dict_load(align_dir)
        all_keylist.extend(kl3)

    info_dir = os.path.join(datadir, 'info_dir')
    info_dict, kl2 = info_dict_load(info_dir)
    all_keylist.extend(kl2)
    supfeat_dict = {}
    kl_dict = {}

    for featname in featlist:
        feat_dir = os.path.join(datadir, featname+'_dir')
        supfeat_dict[featname], kl_dict[featname] =  feat_dict_load(feat_dir, featname)
        all_keylist.append('featname_'+featname)
        all_keylist.extend(kl_dict[featname])
    
    out_dict = {}
    intersection = wav_info_dict.keys() & info_dict.keys() 
    if align_flag:
        intersection = intersection & align_dict.keys() 
    for featname in featlist:
        intersection = intersection & supfeat_dict[featname].keys() 
    
    for utt in intersection:
        outinfo_list = wav_info_dict[utt]
        outinfo_list.extend(info_dict[utt])
        if align_flag:
            outinfo_list.extend(align_dict[utt])
        for featname in featlist:
            outinfo_list.append(featname)
            outinfo_list.extend(supfeat_dict[featname][utt])
        if check_func(outinfo_list):
            out_dict[utt] = outinfo_list
    return out_dict, all_keylist


def check_func(outinfo_list):
    return True

def gen_datalist(supdir_list, outdir, featlist, check_func, prefix, subnum=50000):
    os.makedirs(outdir, exist_ok=True)
    allsave_flag = True
    if allsave_flag:
        all_list_file = os.path.join(outdir, 'all_usage_utt.list')
    sub_table_dir = os.path.join(outdir, 'subtable')
    os.makedirs(sub_table_dir, exist_ok=True)
    spk_id_file = os.path.join(outdir, 'spk_name.dict') 
    list_list_file = os.path.join(outdir, 'list_subtable.list')
    keys_file = os.path.join(outdir, 'keys_name') 

    spkdict = {}
    spkid = 0
    subtabel_id = 0
    tmp_idx = 0
    olistsub = open(list_list_file, 'w')
    if allsave_flag:
        oallt = open(all_list_file, 'w')
    for datadir in supdir_list:
        #out_dict, all_keylist = datadirProcess(datadir, featlist, check_func)
        out_dict, all_keylist = datadirProcess_align(datadir, featlist, check_func)
        for utt in out_dict.keys():
            if tmp_idx >=subnum:
                tmp_idx = 0
                subtabel_id += 1 
                opsubt.close()
            if tmp_idx == 0:
                subtable_file = os.path.join(sub_table_dir, prefix+'_sub_'+str(subtabel_id).rjust(6,'0')+ '.table')  
                opsubt = open(subtable_file, 'w')
                olistsub.write(subtable_file+'\n')
            
            outlist = [utt]
            outlist.extend(out_dict[utt])
            spk = outlist[4]
            if spk not in spkdict.keys():
                spkdict[spk] = spkid
                spkid += 1
            outlist[4] = str(spkid)
            
            outline = '|'.join(outlist) + '\n'
            opsubt.write(outline)
            if allsave_flag:
                oallt.write(outline)

            tmp_idx += 1
    with open(keys_file, 'w') as okf:
        save_keys = ['uttid']
        save_keys.extend(all_keylist)
        oline = '|'.join(save_keys) + '\n'
        okf.write(oline)

    if allsave_flag:
        oallt.close()
    opsubt.close()
    olistsub.close()
    with open(spk_id_file, 'w', encoding="utf-8") as opf:
        json.dump(spkdict, opf, indent=2, ensure_ascii=False)
    
        



'''
FilePath: /python-Dodrio/dodrio/core/datainfo_process.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 15:18:58
LastEditors: Yixiang Chen
LastEditTime: 2025-04-03 17:36:38
'''

import os
import pandas as pd

from dodrio.core.load_info import load_info_dict_libritts, load_info_dict_emilia, load_info_dict_tfile_supdir, load_info_dict_genshin, load_info_dict_table, load_info_dict_single_spk

BLANK_STRING='BlankNone'
info_loadfn_dict = {
    'libritts': load_info_dict_libritts,
    'emilia': load_info_dict_emilia,
    'supdir': load_info_dict_tfile_supdir,
    'genshin': load_info_dict_genshin,
    'table': load_info_dict_table,
    'singlespk': load_info_dict_single_spk,
}

################# other info save ####################

def load_pack_dict(input_dir, from_type='parquet'):
    if from_type == 'parquet':
        load_file = os.path.join(input_dir, 'utt2parquet.list')
    else:
        load_file = os.path.join(input_dir, 'uttinfo.list')
    pack_dict = {}
    with open(load_file, 'r') as lf:
        lines = lf.readlines()
    for line in lines:
        spl = line.strip().split('|')
        name= spl[0]
        packname = spl[1].split('.')[0]
        pack_dict.setdefault(packname, []).append(name)
    return pack_dict

def fake_func(info_dir, kl):
    return 0

def gen_infodir(input_dir, info_dir, out_dir, info_type, kl=['text'], lang='nolang', from_type='parquet', info_func=fake_func):
    '''
        descripttion: 
        param {*} input_dir: data dir 
        param {*} info_dir: text and info original dir
        param {*} out_dir: out_dir
        param {*} info_typ
        param {*} kl
        param {*} lang
        param {*} from_type
        return {*}
    '''
    os.makedirs(out_dir, exist_ok=True)
    pack_dict = load_pack_dict(input_dir, from_type)
    if info_type in info_loadfn_dict.keys():
        fun_load_info = info_loadfn_dict[info_type]
    else:
        fun_load_info = info_func 
    info_dict, keys_list = fun_load_info(info_dir, kl) 
    info_list_file = os.path.join(out_dir, 'uttinfo_text.list')
    opof = open(info_list_file, 'w')
    if (lang != 'nolang') and ('language' not in keys_list):
        outline = '|'.join(keys_list) + '|package_id|language\n'
    else:
        outline = '|'.join(keys_list) + '|package_id\n'
    opof.write(outline)
    for packid in pack_dict.keys():
        uttlist = pack_dict[packid]
        out_df_path = os.path.join(out_dir, packid+'.info')
        df = pd.DataFrame()
        for kk in keys_list:
            #keydata = [info_dict[utt][kk] for utt in uttlist]
            keydata = []
            for utt in uttlist:
                if utt in info_dict.keys():
                    keydata.append(info_dict[utt][kk])
                else:
                    keydata.append(None)
                    print(f'There is no {utt} info for {kk}')
            df[kk] = keydata
        if (lang != 'nolang') and ('language' not in keys_list):
            df['language'] = [lang] * len(uttlist) 
        df.to_parquet(out_df_path)
        print(f'{packid}.info had been Saved')
        for utt in uttlist:
            if utt in info_dict.keys():
                outlist = [str(info_dict[utt][kk]) if info_dict[utt][kk] else BLANK_STRING for kk in keys_list]
                outlist.append(str(packid+'.info'))
                if (lang != 'nolang') and ('language' not in keys_list):
                    outlist.append(lang)
                outline = '|'.join(outlist) + '\n'
                opof.write(outline)
    opof.close()



'''
FilePath: /python-Dodrio/dodrio/afeat/feat_extractor.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-26 19:12:18
LastEditors: Yixiang Chen
LastEditTime: 2025-04-30 10:58:03
'''


import os
from tqdm import tqdm
import numpy as np
import pandas as pd

from dodrio.tools.load_data import load_data_dict, load_pack_audio_data

from dodrio.afeat.exp_load import align_jsondict, align_jsondict_aa, align_jsondict_exc


def get_file_list(inp_dir, suffix='.wav'):
    itm = []
    for home, dirs, files in os.walk(inp_dir):
        pppp = list( map(lambda fname: home + '/' + fname,
            list( filter( lambda filename: os.path.splitext(filename)[1] == suffix,
            files) ) ) )
        itm.extend(pppp)
    file_list = itm
    return file_list

def extract_feat_allload(extractor_func, featname, input_dir, out_dir, from_type, **params):
    os.makedirs(out_dir, exist_ok=True)
    data_dict = load_data_dict(input_dir, from_type)
    feat_info_file = os.path.join(out_dir, 'feat_info_'+featname+'.list')
    oinfo_out = open(feat_info_file, 'w')
    for packid in data_dict.keys():
        uttdict = data_dict[packid]
        out_feat_path = os.path.join(out_dir, packid+'.'+featname)
        outf = open(out_feat_path, 'wb')
        position = 0
        for utt in tqdm(uttdict.keys(), desc=f'{packid} Processing'): 
            wavdata = uttdict[utt]
            feat = extractor_func(wavdata, utt, **params)
            if feat is None:
                feat = np.array([0]).astype(np.float32) # shape is 1
            feat = feat.astype(np.float32)
            fshape = feat.shape
            feat = np.reshape(feat, -1)

            byte_feat = bytes(feat)
            outf.write(byte_feat)

            byte_num = len(feat)* 4 # float 32 = 4 byte 
            end_position = position+byte_num 
            feat_info = [utt, os.path.split(out_feat_path)[-1], str(position), str(end_position), ','.join([str(xx) for xx in fshape])]
            info_outline = '|'.join(feat_info) + '\n'
            oinfo_out.write(info_outline) 

            position += byte_num
        outf.close()
    oinfo_out.close()

def extract_feat(extractor_func, featname, input_dir, out_dir, from_type, **params):
    os.makedirs(out_dir, exist_ok=True)
    #data_dict = load_data_dict(input_dir, from_type)
    if from_type=='parquet':
        suffix = '.parquet'
        infolistf = ''
    else:
        suffix = '.pack'
        infolistf = os.path.join(input_dir, 'uttinfo.list')
    packlist = get_file_list(input_dir, suffix)
    packlist.sort()
    feat_info_file = os.path.join(out_dir, 'feat_info_'+featname+'.list')
    oinfo_out = open(feat_info_file, 'w')

    for idx in range(len(packlist)):
        packfile = packlist[idx]
        packid = os.path.split(packfile)[-1].split(suffix)[0]
        uttdict = load_pack_audio_data(packfile, infolistf, return_sr = False)
        
        out_feat_path = os.path.join(out_dir, packid+'.'+featname)
        outf = open(out_feat_path, 'wb')
        position = 0
        for utt in tqdm(uttdict.keys(), desc=f'{packid} Processing'): 
            wavdata = uttdict[utt]
            feat = extractor_func(wavdata, utt, **params)
            if feat is None:
                feat = np.array([0]).astype(np.float32) # shape is 1
            feat = feat.astype(np.float32)
            fshape = feat.shape
            feat = np.reshape(feat, -1)

            byte_feat = bytes(feat)
            outf.write(byte_feat)

            byte_num = len(feat)* 4 # float 32 = 4 byte 
            end_position = position+byte_num 
            feat_info = [utt, os.path.split(out_feat_path)[-1], str(position), str(end_position), ','.join([str(xx) for xx in fshape])]
            info_outline = '|'.join(feat_info) + '\n'
            oinfo_out.write(info_outline) 

            position += byte_num
        outf.close()
    oinfo_out.close()

def extract_feat_extrainfo(extractor_func, featname, input_dir, out_dir, from_type, **params):
    os.makedirs(out_dir, exist_ok=True)
    #data_dict = load_data_dict(input_dir, from_type)
    if from_type=='parquet':
        suffix = '.parquet'
        infolistf = ''
    else:
        suffix = '.pack'
        infolistf = os.path.join(input_dir, 'uttinfo.list')
    packlist = get_file_list(input_dir, suffix)
    packlist.sort()
    feat_info_file = os.path.join(out_dir, 'feat_info_'+featname+'.list')
    oinfo_out = open(feat_info_file, 'w')

    for idx in range(len(packlist)):
        packfile = packlist[idx]
        packid = os.path.split(packfile)[-1].split(suffix)[0]
        uttdict = load_pack_audio_data(packfile, infolistf, return_sr = False)
        
        out_feat_path = os.path.join(out_dir, packid+'.'+featname)
        outf = open(out_feat_path, 'wb')
        position = 0
        for utt in tqdm(uttdict.keys(), desc=f'{packid} Processing'): 
            wavdata = uttdict[utt]
            try:
                feat, extrainfo = extractor_func(wavdata, utt, **params)
            except:
                feat = None
                extrainfo = ['0']
            if feat is None:
                feat = np.array([0]).astype(np.float32) # shape is 1
            feat = feat.astype(np.float32)
            fshape = feat.shape
            feat = np.reshape(feat, -1)

            byte_feat = bytes(feat)
            outf.write(byte_feat)

            byte_num = len(feat)* 4 # float 32 = 4 byte 
            end_position = position+byte_num 
            feat_info = [utt, os.path.split(out_feat_path)[-1], str(position), str(end_position), ','.join([str(xx) for xx in fshape])]
            feat_info.extend(extrainfo)
            info_outline = '|'.join(feat_info) + '\n'
            oinfo_out.write(info_outline) 

            position += byte_num
        outf.close()
    oinfo_out.close()


def save_parquet_align(wavinfo_dict, savelist, parquet_fn):
    word_start = [wavinfo_dict[x][0] for x in savelist]
    word_end = [wavinfo_dict[x][1] for x in savelist]
    word_list = [wavinfo_dict[x][2] for x in savelist]

    phone_start = [wavinfo_dict[x][3] for x in savelist]
    phone_end = [wavinfo_dict[x][4] for x in savelist]
    phone_list = [wavinfo_dict[x][5] for x in savelist]
    phone_duration = [wavinfo_dict[x][6] for x in savelist] 
    df = pd.DataFrame()
    df['utt'] = savelist
    df['word_start'] = word_start
    df['word_end'] = word_end
    df['word_list'] = word_list
    df['phone_start'] = phone_start
    df['phone_end'] = phone_end
    df['phone_list'] = phone_list
    df['phone_duration'] = phone_duration
    df.to_parquet(parquet_fn)
    #print(f"{parquet_fn} had be saved")

def extract_feat_align(extractor_func, featname, input_dir, out_dir, from_type, jsondir, **params):
    os.makedirs(out_dir, exist_ok=True)
    #data_dict = load_data_dict(input_dir, from_type)
    if from_type=='parquet':
        suffix = '.parquet'
        infolistf = ''
    else:
        suffix = '.pack'
        infolistf = os.path.join(input_dir, 'uttinfo.list')
    packlist = get_file_list(input_dir, suffix)
    packlist.sort()
    feat_info_file = os.path.join(out_dir, 'feat_info_'+featname+'.list')
    oinfo_out = open(feat_info_file, 'w')

    jsonfiledict = align_jsondict_aa(jsondir)
    #jsonfiledict = align_jsondict_exc(jsondir)

    for idx in range(len(packlist)):
        packfile = packlist[idx]
        packid = os.path.split(packfile)[-1].split(suffix)[0]
        uttdict = load_pack_audio_data(packfile, infolistf, return_sr = False)
        
        out_feat_path = os.path.join(out_dir, packid+'.'+featname)
        info_dict = {}
        for utt in tqdm(uttdict.keys(), desc=f'{packid} Processing'): 
            wavdata = uttdict[utt]
            if utt not in jsonfiledict.keys():
                continue
            wstart, wend, wlist, pstart, pend, plist, pdur = extractor_func(wavdata, utt, jsonfiledict, **params)

            info_dict[utt] = [wstart, wend, wlist, pstart, pend, plist, pdur]

            feat_info = [utt, os.path.split(out_feat_path)[-1], ' '.join(plist), ' '.join([str(xx) for xx in pdur]) ]
            info_outline = '|'.join(feat_info) + '\n'
            oinfo_out.write(info_outline) 

        save_parquet_align(info_dict, info_dict.keys(), out_feat_path)
    oinfo_out.close()

def extract_feat_multi(extractor_func, featname_list, input_dir, out_dir_list, from_type, **params):

    if from_type=='parquet':
        suffix = '.parquet'
        infolistf = ''
    else:
        suffix = '.pack'
        infolistf = os.path.join(input_dir, 'uttinfo.list')
    packlist = get_file_list(input_dir, suffix)
    packlist.sort()

    oinfo_out_list = []
    for ii in range(len(featname_list)):
        out_dir = out_dir_list[ii]
        featname = featname_list[ii]
        os.makedirs(out_dir, exist_ok=True)
        feat_info_file = os.path.join(out_dir, 'feat_info_'+featname+'.list')
        oinfo_out = open(feat_info_file, 'w')
        oinfo_out_list.append(oinfo_out)

    for idx in range(len(packlist)):
        packfile = packlist[idx]
        packid = os.path.split(packfile)[-1].split(suffix)[0]
        uttdict = load_pack_audio_data(packfile, infolistf, return_sr = False)
        
        positions = []
        outf_list = []
        out_feat_path_list = []
        for ii in range(len(featname_list)):
            positions.append(0)
            featname = featname_list[ii]
            out_dir = out_dir_list[ii]
            out_feat_path = os.path.join(out_dir, packid+'.'+featname)
            out_feat_path_list.append(out_feat_path)
            outf = open(out_feat_path, 'wb')
            outf_list.append(outf)

        for utt in tqdm(uttdict.keys(), desc=f'{packid} Processing'): 
            wavdata = uttdict[utt]
            feat_list = extractor_func(wavdata, utt, **params)
            for ii in range(len(feat_list)):
                feat = feat_list[ii]
                if feat is None:
                    feat = np.array([0]).astype(np.float32) # shape is 1
                feat = feat.astype(np.float32)
                fshape = feat.shape
                feat = np.reshape(feat, -1)

                byte_feat = bytes(feat)
                outf_list[ii].write(byte_feat)

                byte_num = len(feat)* 4 # float 32 = 4 byte 
                end_position = positions[ii]+byte_num 
                #feat_info = [utt, os.path.split(out_feat_path)[-1], str(positions[ii]), str(end_position), ','.join([str(xx) for xx in fshape])]
                feat_info = [utt, os.path.split(out_feat_path_list[ii])[-1], str(positions[ii]), str(end_position), ','.join([str(xx) for xx in fshape])]
                info_outline = '|'.join(feat_info) + '\n'
                oinfo_out_list[ii].write(info_outline) 

                positions[ii] += byte_num
        for outf in outf_list:
            outf.close()
    for oinfo_out in oinfo_out_list:
        oinfo_out.close()


def get_utt2spk(infodir):
    utt2spk = {}
    inpfile = os.path.join(infodir, 'uttinfo_text.list')
    with open(inpfile, 'r') as oif:
        info_lines = oif.readlines()
    index_list = info_lines[0].strip().split('|')
    idid, spkid = -1, -1 
    for idx in range(len(index_list)):
        if index_list[idx] == 'id':
            idid = idx
        if index_list[idx] == 'speaker':
            spkid = idx
    for idx in range(1, len(info_lines)):
        spl = info_lines[idx].strip().split('|')
        utt = spl[idid]
        spk = spl[spkid]
        utt2spk[utt] = spk
    del info_lines
    return utt2spk





'''
FilePath: /python-Dodrio/dodrio/afeat/exp_load.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-04-14 16:56:12
LastEditors: Yixiang Chen
LastEditTime: 2025-04-15 15:19:07
'''

import json
import os

def get_file_list(inp_dir, suffix='.wav'):
    itm = []
    for home, dirs, files in os.walk(inp_dir):
        pppp = list( map(lambda fname: home + '/' + fname,
            list( filter( lambda filename: os.path.splitext(filename)[1] == suffix,
            files) ) ) )
        itm.extend(pppp)
    file_list = itm
    return file_list

def load_mfajson(jsonfile):
    with open(jsonfile, 'r') as af:
        adict = json.load(af)
    word_align_list = adict['tiers']['words']['entries']
    phone_align_list = adict['tiers']['phones']['entries']

    wstart = [ww[0]*1000 for ww in word_align_list]
    wend = [ww[1]*1000 for ww in word_align_list] 
    wlist = [ww[2] for ww in word_align_list]

    pstart = [pp[0]*1000 for pp in phone_align_list]
    pend = [pp[1]*1000 for pp in phone_align_list]
    plist = [pp[2] for pp in phone_align_list]

    pdur = [int(int(pend[idx])-int(pstart[idx])) for idx in range(len(pstart))]

    return wstart, wend, wlist, pstart, pend, plist, pdur

def get_mfa(wav, utt, filedict):
    jsonfile = filedict[utt]
    wstart, wend, wlist, pstart, pend, plist, pdur = load_mfajson(jsonfile)
    return wstart, wend, wlist, pstart, pend, plist, pdur 

def align_jsondict(jsondir):
    jsonflist = get_file_list(jsondir, '.json')
    jsonfiledict = {}
    for jsonf in jsonflist:
        basename = os.path.split(jsonf)[-1].split('.json')[0]
        jsonfiledict[basename] = jsonf
    return jsonfiledict 

def align_jsondict_aa(jsondir):
    jsonfiledict = {}
    eplist = os.listdir(jsondir)
    for epd in eplist:
        aligndir = os.path.join(jsondir, epd, 'align_new')
        jsonflist = get_file_list(aligndir, '.json')
        for jsonf in jsonflist:
            basename = os.path.split(jsonf)[-1].split('.json')[0]
            jsonfiledict[basename] = jsonf
    return jsonfiledict

def align_jsondict_exc(jsondir):
    jsonfiledict = {}
    aligndir = os.path.join(jsondir, 'align_new')
    jsonflist = get_file_list(aligndir, '.json')
    for jsonf in jsonflist:
        basename = os.path.split(jsonf)[-1].split('.json')[0]
        basename = '_'.join(basename.split('_')[1:])
        jsonfiledict[basename] = jsonf
    return jsonfiledict



'''
FilePath: /python-Dodrio/dodrio/__init__.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 15:50:51
LastEditors: Yixiang Chen
LastEditTime: 2025-04-29 16:36:03
'''

#from . import tool
#from . import genlist 
#from . import afeat


from dodrio.core.parquet_base import (
    gen_parquet, parquet2wav
)

from dodrio.core.package_base import (
    parquet2package, package2wav, gen_package
)

from dodrio.core.datainfo_process import gen_infodir

from dodrio.genlist.gendatalist import (
    genListDir, gen_datalist, check_func
)

from dodrio.afeat.feat_extractor import (
    extract_feat, get_utt2spk, extract_feat_multi, extract_feat_align, extract_feat_extrainfo
)

from dodrio.tools.load_data import (
    load_data_from_line, load_feat_single, load_audio_single, load_data_from_line_align
)



'''
FilePath: /python-Dodrio/setup.py
Descripttion: 
Author: Yixiang Chen
version: 
Date: 2025-03-24 15:57:41
LastEditors: Yixiang Chen
LastEditTime: 2025-05-12 19:48:23
'''

from setuptools import setup, find_packages

VERSION = '0.3.4' 
DESCRIPTION = 'Data Package for TTS '
#LONG_DESCRIPTION = 'Data Package for TTS with a slightly longer description'
LONG_DESCRIPTION = open("README.md", encoding="utf-8").read()

# 配置
setup(
        name="dodrio", 
        version=VERSION,
        author="Yixiang Chen",
        author_email="<yixiangchen1995@gmail.com>",
        license='MIT',
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        url="https://github.com/yixiangchen1995/python-Dodrio",
        packages=find_packages(),
        install_requires=[
            'numpy',
            'pandas',
            'pyarrow',
            'scipy',
            'tqdm',
            'librosa'
        ], # add any additional packages that ## add tinytag package
        
        keywords=['python', 'first package'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Programming Language :: Python :: 3.13',
        ]
)



import os
import numpy as np
import librosa
from tqdm import tqdm
import glob
from io import BytesIO
import json
import math
from scipy.io import wavfile
import pyarrow.parquet as pq

from dodrio.core.utils import set_wavlist

############## Parquet to Package ###############   

def audio_regular(audio, ori_sample_rate, dtype, target_sample_rate=48000):
    '''
    descripttion: 
    param {*} audio
    param {*} ori_sample_rate
    param {str} dtype : 'int16' or 'int32'
    param {int} target_sample_rate : target sample rate
    return {*}
    '''
    if dtype == 'int16':
        dnum = 32768
    elif dtype == 'int32':
        dnum = 2147483648
    else:
        print(f"{dtype} is not normal data type")
    norm_audio = audio / dnum
    rs_audio = librosa.resample(norm_audio, orig_sr=ori_sample_rate, target_sr=target_sample_rate)
    int16_audio = (rs_audio*32768).astype(np.int16)
    return int16_audio

def load_mp3_frombio(bio, target_sample_rate=48000):
    audio, ori_sr = librosa.load(BytesIO(bio))
    rs_audio = librosa.resample(audio, orig_sr=ori_sr, target_sr=target_sample_rate)
    int16_audio = (rs_audio*32768).astype(np.int16)
    return int16_audio

def save_meta_info(outdir, sample_rate, num_utts_per_parquet):
    outpath = os.path.join(outdir, 'meta_info.json')
    outinfo = {
        'sample_rate': sample_rate,
        'num_utts_per_parquet': num_utts_per_parquet,
    }
    with open(outpath, 'w', encoding="utf-8") as opf:
        json.dump(outinfo, opf, indent=2, ensure_ascii=False)

def parquet2package_single(parquet_file, package_file, target_sample_rate=48000):
    ftype = os.path.split(parquet_file)[-1].split('_')[0]
    df = pq.read_table(parquet_file).to_pandas()
    basename = os.path.split(parquet_file)[-1].split('.parquet')[0]
    position = 0
    info_list = []
    audio_pos = {}
    outf = open(package_file, 'wb')
    num_utts_per_parquet = len(df)
    for idx in tqdm(range(len(df)), desc=f'{basename} Processing'):
        utt = df.iloc[idx]['utt']
        sr = df.iloc[idx]['sample_rate']
        dtype = df.iloc[idx]['dtype']
        audio = df.iloc[idx]['audio_data']
        if ftype == 'wav':
            reg_audio = audio_regular(audio, sr, dtype, target_sample_rate)
        elif ftype == 'mp3':
            reg_audio = load_mp3_frombio(audio, target_sample_rate) 
        else:
            print("Now just accept mp3 and wav format")
            return

        byte_audio = bytes(reg_audio)
        outf.write(byte_audio)

        byte_num = len(reg_audio)* 2 
        end_position = position+byte_num 
        audio_pos[utt] = [position, end_position]
        info_list.append([utt, os.path.split(package_file)[-1], str(position), str(end_position)])

        position += byte_num
    return info_list, num_utts_per_parquet

def parquet2package(parquet_dir, package_dir, sample_rate=48000):
    os.makedirs(package_dir, exist_ok=True)
    pq_file_list = glob.glob(parquet_dir+'/*.parquet')
    pq_file_list.sort()
    all_info_list = []
    sflag = True
    for one_pq in pq_file_list:
        basename = os.path.split(one_pq)[-1].split('.parquet')[0]
        ftype = os.path.split(one_pq)[-1].split('_')[0]
        pack_file = os.path.join(package_dir, basename+'.pack')
        info_list, num_utts_per_parquet = parquet2package_single(one_pq, pack_file, sample_rate)
        if sflag:
            save_meta_info(package_dir, sample_rate, num_utts_per_parquet)
            sflag = False
        all_info_list.append(info_list)
    info_outfile = os.path.join(package_dir, 'uttinfo.list')
    with open(info_outfile, 'w') as outf:
        for info_list in all_info_list:
            for info in info_list:
                outline = '|'.join(info) + '\n'
                outf.write(outline)

############## Wav to Package ############### 

def load_mp3(mp3path, target_sample_rate=48000):
    audio, ori_sr = librosa.load(mp3path)
    rs_audio = librosa.resample(audio, orig_sr=ori_sr, target_sr=target_sample_rate)
    int16_audio = (rs_audio*32768).astype(np.int16)
    return int16_audio

def load_audio_content(utt, audiopath, target_sample_rate, file_type='wav'):
    if file_type == 'wav':
        sr, audio = wavfile.read(audiopath)
        if len(audio)<1:
            print(f"{utt} wavfile is None")
            return None
        if len(audio.shape) > 1:
            print(f"{utt} {str(len(audio.shape))} channel is non-mono channel, just first channel will be save")
            audio = audio[:,0]
        dtype = str(audio.dtype)
        int16_audio = audio_regular(audio, sr, dtype, target_sample_rate)
    elif file_type == 'mp3':
        int16_audio = load_mp3(audiopath, target_sample_rate) 
    else:
        print("Now just accept mp3 and wav format")
        return None
    return int16_audio 

def gen_package(wav_dir, package_dir, mid_name='', target_sample_rate=48000, file_type='wav', num_utts_per_parquet=2000, process_max_num=10000):
    os.makedirs(package_dir, exist_ok=True)
    wavdict, uttlist = set_wavlist(wav_dir, file_type)

    pack2utt = {}
    turn_num = math.ceil(len(uttlist) / process_max_num)
    all_info_list = []
    true_len = min(len(uttlist), num_utts_per_parquet)
    save_meta_info(package_dir, target_sample_rate, true_len)
    for tid in range(turn_num):
        wavinfo_dict ={}
        for utt in tqdm(uttlist[process_max_num * tid : process_max_num * (tid+1)], desc=f'Turn{str(tid)}LoadAudio'):
            wavpath = wavdict[utt]
            int16_audio = load_audio_content(utt, wavpath, target_sample_rate)
            wavinfo_dict[utt] = int16_audio

        prefix = file_type
        pack_list = []
        for i, j in enumerate(range(process_max_num * tid, min(len(uttlist), process_max_num * (tid+1)), num_utts_per_parquet)):
            idx = (process_max_num * tid) // num_utts_per_parquet + i
            pfile = prefix + '_' + mid_name + '_{:05d}.pack'.format(idx)
            pack_file = os.path.join(package_dir, pfile)
            pack_list.append(pack_file)
            pack2utt[pfile] = uttlist[j: j + num_utts_per_parquet] 

            position = 0
            info_list = []
            audio_pos = {}
            outf = open(pack_file, 'wb')
            for idx in tqdm(range(len(uttlist[j: j + num_utts_per_parquet])), desc=f'{pfile} Processing'):
                utt = uttlist[j+idx]
                int16_audio = wavinfo_dict[utt]
                if int16_audio is None:
                    continue
                byte_audio = bytes(int16_audio)
                outf.write(byte_audio)

                byte_num = len(int16_audio)* 2 
                end_position = position+byte_num 
                audio_pos[utt] = [position, end_position]
                info_list.append([utt, os.path.split(pack_file)[-1], str(position), str(end_position)])

                position += byte_num
            outf.close()
            all_info_list.append(info_list)
    info_outfile = os.path.join(package_dir, 'uttinfo.list')
    with open(info_outfile, 'w') as outf:
        for info_list in all_info_list:
            for info in info_list:
                outline = '|'.join(info) + '\n'
                outf.write(outline)

############## Package to Wav ############### 

def load_data_from_package(pack_file, start, end):
    with open(pack_file, 'rb') as pf :
        pf.seek(start)
        data = pf.read(end-start)
    audio = np.frombuffer(data, dtype=np.int16)
    return audio
def save_wav(audio, wavpath, sr):
    wavfile.write(wavpath, sr, audio)
    
def package2wav(package_dir, wav_dir):
    os.makedirs(wav_dir, exist_ok=True)
    info_file = os.path.join(package_dir, 'uttinfo.list')
    meta_file = os.path.join(package_dir, 'meta_info.json')
    with open(meta_file, 'r') as fm:
        minfo = json.load(fm)
        sample_rate = minfo['sample_rate']
    with open(info_file, 'r') as iff:
        info_list = iff.readlines()
    for info in tqdm(info_list):
        utt, pf, start, end = info.split('|')
        start = int(start)
        end = int(end)
        wpf = os.path.join(package_dir, pf)
        audio = load_data_from_package(wpf, start, end)
        wavpath = os.path.join(wav_dir, utt+'.wav')
        save_wav(audio, wavpath, sample_rate)